﻿' ActiveVB DHCP Discovery
' Converted from  http://www.indented.co.uk/2010/02/17/dhcp-discovery/
' Author: Frank Schüler

'<#
'  Net-DhcpDiscover.ps1

'  Author: Chris Dent
'  Date :  16/02/2010

'  A script To send a DHCPDISCOVER request And report On DHCPOFFER 
'  responses returned by all DHCP Servers On the current subnet.

'  DHCP Packet Format (RFC 2131 - http://www.ietf.org/rfc/rfc2131.txt)

'  0                   1                   2                   3
'  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
'  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'  |     op (1)    |   htype (1)   |   hlen (1)    |   hops (1)    |
'  +---------------+---------------+---------------+---------------+
'  |                            xid (4)                            |
'  +-------------------------------+-------------------------------+
'  |           secs (2)            |           flags (2)           |
'  +-------------------------------+-------------------------------+
'  |                          ciaddr  (4)                          |
'  +---------------------------------------------------------------+
'  |                          yiaddr  (4)                          |
'  +---------------------------------------------------------------+
'  |                          siaddr  (4)                          |
'  +---------------------------------------------------------------+
'  |                          giaddr  (4)                          |
'  +---------------------------------------------------------------+
'  |                                                               |
'  |                          chaddr  (16)                         |
'  |                                                               |
'  |                                                               |
'  +---------------------------------------------------------------+
'  |                                                               |
'  |                          sname   (64)                         |
'  +---------------------------------------------------------------+
'  |                                                               |
'  |                          file    (128)                        |
'  +---------------------------------------------------------------+
'  |                                                               |
'  |                          options (variable)                   |
'  +---------------------------------------------------------------+

'   FIELD      OCTETS       DESCRIPTION
'   -----      ------       -----------

'   op            1  Message op code / message type.
'                    1 = BOOTREQUEST, 2 = BOOTREPLY
'   htype         1  Hardware address type, see ARP section In "Assigned
'                    Numbers" RFC; e.g., '1' = 10mb ethernet.
'   hlen          1  Hardware address length (e.g.  '6' for 10mb
'                    ethernet).
'   hops          1  Client sets To zero, optionally used by relay agents
'                    when booting via a relay agent.
'   xid           4  Transaction ID, a random number chosen by the
'                    client, used by the client And server to associate
'                    messages And responses between a client And a
'                    server.
'   secs          2  Filled In by client, seconds elapsed since client
'                    began address acquisition Or renewal process.
'   flags         2  Flags (see figure 2).
'   ciaddr        4  Client IP address; only filled in if client Is in
'                    BOUND, RENEW Or REBINDING state And can respond
'                    to ARP requests.
'   yiaddr        4  'your' (client) IP address.
'   siaddr        4  IP address of next server to use in bootstrap;
'                    returned in DHCPOFFER, DHCPACK by server.
'   giaddr        4  Relay agent IP address, used in booting via a
'                    relay agent.
'   chaddr       16  Client hardware address.
'   sname        64  Optional server host name, null terminated string.
'   file        128  Boot file name, null terminated string; "generic"
'                    name Or null In DHCPDISCOVER, fully qualified
'                    directory-path name in DHCPOFFER.
'   options     var  Optional parameters field.  See the options
'                    documents for a list of defined options.
'#>

Option Strict On
Option Explicit On
Imports System.Text.RegularExpressions

Public Class Form1

    Private Structure DhcpPacketOption
        Dim OptionCode As Byte
        Dim OptionName As String
        Dim Length As Byte
        Dim OptionValue As String
    End Structure

    Private Structure DhcpPacket
        Dim Op As String
        Dim HType As String
        Dim HLen As Byte
        Dim Hops As Byte
        Dim XID As UInt32
        Dim Secs As UInt16
        Dim Flags As String
        Dim CIAddr As String
        Dim YIAddr As String
        Dim SIAddr As String
        Dim GIAddr As String
        Dim CHAddr As String
        Dim SName As String
        Dim File As String
        Dim MagicCookie As String
        Dim Options As DhcpPacketOption()
    End Structure

    ' Length Of time (In seconds) To spend waiting For Offers If
    ' the connection does Not timeout first
    Private DiscoverTimeout As Byte = 60

    ' Build a DHCPDISCOVER packet to send
    Private Function New_DhcpDiscoverPacket(Optional ByVal MacAddressString As String = "AA:BB:CC:DD:EE:FF") As Byte()

        ' Generate a Transaction ID for this request
        Dim XID As Byte() = New Byte(3) {}
        Dim Random = New Random
        Random.NextBytes(XID)

        ' Convert the MAC Address String into a Byte Array
        ' Drop any characters which might be used to delimit the string
        MacAddressString = Regex.Replace(MacAddressString, "[^a-fA-F0-9]", "")
        Dim MacAddress As Byte() = BitConverter.GetBytes((UInt64.Parse(MacAddressString, Globalization.NumberStyles.HexNumber)))
        ' Reverse the MAC Address array
        Array.Reverse(MacAddress)

        ' Create the Byte Array
        Dim DhcpDiscover As Byte() = New Byte(242) {}

        ' Copy the Transaction ID Bytes into the array
        Array.Copy(XID, 0, DhcpDiscover, 4, 4)

        ' Copy the MacAddress Bytes into the array (drop the first 2 bytes, 
        ' too many bytes returned from UInt64)
        Array.Copy(MacAddress, 2, DhcpDiscover, 28, 6)

        ' Set the OP Code to BOOTREQUEST
        DhcpDiscover(0) = 1
        ' Set the Hardware Address Type to Ethernet
        DhcpDiscover(1) = 1
        ' Set the Hardware Address Length (number of bytes)
        DhcpDiscover(2) = 6
        ' Set the Broadcast Flag
        DhcpDiscover(10) = 128
        ' Set the Magic Cookie values
        DhcpDiscover(236) = 99
        DhcpDiscover(237) = 130
        DhcpDiscover(238) = 83
        DhcpDiscover(239) = 99
        ' Set the DHCPDiscover Message Type Option
        DhcpDiscover(240) = 53
        DhcpDiscover(241) = 1
        DhcpDiscover(242) = 1

        Return DhcpDiscover

    End Function

    ' Parse a DHCP Packet, returning an object containing each field
    Private Function Read_DhcpPacket(ByVal Packet As Byte()) As DhcpPacket

        Dim TmpArr As Byte() = New Byte(3) {}
        Dim Reader As IO.BinaryReader = New IO.BinaryReader(New IO.MemoryStream(Packet))
        Dim DhcpResponse As DhcpPacket = New DhcpPacket

        ' Get And translate the Op code
        If Reader.ReadByte() = 1 Then
            DhcpResponse.Op = "BootRequest"
        Else
            DhcpResponse.Op = "BootResponse"
        End If
        Debug.Print("DhcpResponse.Op = " & DhcpResponse.Op)

        If Reader.ReadByte() = 1 Then DhcpResponse.HType = "Ethernet"
        Debug.Print("DhcpResponse.HType = " & DhcpResponse.HType)

        DhcpResponse.HLen = Reader.ReadByte()
        Debug.Print("DhcpResponse.HLen = " & DhcpResponse.HLen.ToString)
        DhcpResponse.Hops = Reader.ReadByte()
        Debug.Print("DhcpResponse.Hops = " & DhcpResponse.Hops.ToString)
        DhcpResponse.XID = Reader.ReadUInt32()
        Debug.Print("DhcpResponse.XID = " & DhcpResponse.XID.ToString)
        DhcpResponse.Secs = Reader.ReadUInt16()
        Debug.Print("DhcpResponse.Secs = " & DhcpResponse.Secs.ToString)

        ' Broadcast Is the only flag that can be present, the other bits are reserved
        If CBool(Reader.ReadUInt16 And 128) Then DhcpResponse.Flags = "Broadcast"
        Debug.Print("DhcpResponse.Flags = " & DhcpResponse.Flags)

        Reader.Read(TmpArr, 0, 4)
        DhcpResponse.CIAddr = String.Join(".", TmpArr)
        Debug.Print("DhcpResponse.CIAddr = " & DhcpResponse.CIAddr)

        Reader.Read(TmpArr, 0, 4)
        DhcpResponse.YIAddr = String.Join(".", TmpArr)
        Debug.Print("DhcpResponse.YIAddr = " & DhcpResponse.YIAddr)

        Reader.Read(TmpArr, 0, 4)
        DhcpResponse.SIAddr = String.Join(".", TmpArr)
        Debug.Print("DhcpResponse.SIAddr = " & DhcpResponse.SIAddr)

        Reader.Read(TmpArr, 0, 4)
        DhcpResponse.GIAddr = String.Join(".", TmpArr)
        Debug.Print("DhcpResponse.GIAddr = " & DhcpResponse.GIAddr)

        Dim MacAddrBytes As Byte() = New Byte(15) {}
        Reader.Read(MacAddrBytes, 0, 16)
        Array.Resize(MacAddrBytes, 5)
        DhcpResponse.CHAddr = BitConverter.ToString(MacAddrBytes).Replace("-", ":")
        Debug.Print("DhcpResponse.CHAddr = " & DhcpResponse.CHAddr)

        DhcpResponse.SName = String.Join("", Reader.ReadChars(64)).Trim
        Debug.Print("DhcpResponse.SName = " & DhcpResponse.SName)
        Debug.Print("")

        DhcpResponse.File = String.Join("", Reader.ReadChars(128)).Trim
        Debug.Print("DhcpResponse.File = " & DhcpResponse.File)
        Debug.Print("")

        Reader.Read(TmpArr, 0, 4)
        DhcpResponse.MagicCookie = String.Join(" ", TmpArr)
        Debug.Print("DhcpResponse.MagicCookie = " & DhcpResponse.MagicCookie)

        Dim Count As Int32 = 0

        ' Start reading Options
        While Reader.BaseStream.Position < Reader.BaseStream.Length

            Array.Resize(DhcpResponse.Options, Count + 1)

            DhcpResponse.Options(Count).OptionCode = Reader.ReadByte()
            DhcpResponse.Options(Count).OptionName = ""
            DhcpResponse.Options(Count).Length = 0
            DhcpResponse.Options(Count).OptionValue = ""

            If DhcpResponse.Options(Count).OptionCode <> 0 And DhcpResponse.Options(Count).OptionCode <> 255 Then
                DhcpResponse.Options(Count).Length = Reader.ReadByte()
            End If

            Select Case DhcpResponse.Options(Count).OptionCode
                Case 0
                    DhcpResponse.Options(Count).OptionName = "PadOption"
                Case 1
                    DhcpResponse.Options(Count).OptionName = "SubnetMask"
                    Reader.Read(TmpArr, 0, 4)
                    DhcpResponse.Options(Count).OptionValue = String.Join(".", TmpArr)
                Case 3
                    DhcpResponse.Options(Count).OptionName = "Router"
                    Reader.Read(TmpArr, 0, 4)
                    DhcpResponse.Options(Count).OptionValue = String.Join(".", TmpArr)
                Case 6
                    DhcpResponse.Options(Count).OptionName = "DomainNameServer"
                    For i = 0 To DhcpResponse.Options(Count).Length / 4
                        Reader.Read(TmpArr, 0, 4)
                        If i < DhcpResponse.Options(Count).Length / 4 Then
                            DhcpResponse.Options(Count).OptionValue &= String.Join(".", TmpArr) & " / "
                        Else
                            DhcpResponse.Options(Count).OptionValue &= String.Join(".", TmpArr)
                        End If
                    Next
                Case 15
                    DhcpResponse.Options(Count).OptionName = "DomainName"
                    DhcpResponse.Options(Count).OptionValue = String.Join("", Reader.ReadChars(DhcpResponse.Options(Count).Length))
                Case 51
                    DhcpResponse.Options(Count).OptionName = "IPAddressLeaseTime"
                    ' Read as Big Endian
                    Dim Value = (Reader.ReadByte() * Math.Pow(256, 3)) +
                        (Reader.ReadByte() * Math.Pow(256, 2)) +
                        (Reader.ReadByte() * 256) + Reader.ReadByte
                    DhcpResponse.Options(Count).OptionValue = TimeSpan.FromSeconds(Value).ToString
                Case 53
                    DhcpResponse.Options(Count).OptionName = "DhcpMessageType"
                    Select Case Reader.ReadByte()
                        Case 1
                            DhcpResponse.Options(Count).OptionValue = "DHCPDISCOVER"
                        Case 2
                            DhcpResponse.Options(Count).OptionValue = "DHCPOFFER"
                        Case 3
                            DhcpResponse.Options(Count).OptionValue = "DHCPREQUEST"
                        Case 4
                            DhcpResponse.Options(Count).OptionValue = "DHCPDECLINE"
                        Case 5
                            DhcpResponse.Options(Count).OptionValue = "DHCPACK"
                        Case 6
                            DhcpResponse.Options(Count).OptionValue = "DHCPNAK"
                        Case 7
                            DhcpResponse.Options(Count).OptionValue = "DHCPRELEASE"
                    End Select
                Case 54
                    DhcpResponse.Options(Count).OptionName = "DhcpServerIdentifier"
                    Reader.Read(TmpArr, 0, 4)
                    DhcpResponse.Options(Count).OptionValue = String.Join(".", TmpArr)
                Case 58
                    DhcpResponse.Options(Count).OptionName = "RenewalTime"
                    ' Read as Big Endian
                    Dim Value = (Reader.ReadByte() * Math.Pow(256, 3)) +
                        (Reader.ReadByte() * Math.Pow(256, 2)) +
                        (Reader.ReadByte() * 256) + Reader.ReadByte
                    DhcpResponse.Options(Count).OptionValue = TimeSpan.FromSeconds(Value).ToString
                Case 59
                    DhcpResponse.Options(Count).OptionName = "RebindingTime"
                    ' Read as Big Endian
                    Dim Value = (Reader.ReadByte() * Math.Pow(256, 3)) +
                        (Reader.ReadByte() * Math.Pow(256, 2)) +
                        (Reader.ReadByte() * 256) + Reader.ReadByte
                    DhcpResponse.Options(Count).OptionValue = TimeSpan.FromSeconds(Value).ToString
                Case 255
                    DhcpResponse.Options(Count).OptionName = "EndOption"
                Case Else
                    ' For all options which are not decoded here
                    DhcpResponse.Options(Count).OptionName = "NoOptionDecode"
                    Dim Buffer As Byte() = New Byte(DhcpResponse.Options(Count).Length - 1) {}
                    Reader.Read(Buffer, 0, DhcpResponse.Options(Count).Length)
                    DhcpResponse.Options(Count).OptionValue = String.Join(" ", Buffer)
            End Select

            If DhcpResponse.Options(Count).OptionCode <> 0 Then
                Debug.Print(DhcpResponse.Options(Count).OptionName & " = " & DhcpResponse.Options(Count).OptionValue)
            End If

            Count += 1

        End While

        Return DhcpResponse

    End Function


    ' Create a UDP Socket with Broadcast and Address Re-use enabled.
    Private Function New_UdpSocket(Optional ByVal SendTimeOut As Int32 = 5, Optional ByVal ReceiveTimeOut As Int32 = 5) As Net.Sockets.Socket

        Dim UdpSocket As Net.Sockets.Socket = New Net.Sockets.Socket(
            Net.Sockets.AddressFamily.InterNetwork,
            Net.Sockets.SocketType.Dgram,
            Net.Sockets.ProtocolType.Udp)

        UdpSocket.EnableBroadcast = True
        UdpSocket.ExclusiveAddressUse = False
        UdpSocket.SendTimeout = SendTimeOut * 1000
        UdpSocket.ReceiveTimeout = ReceiveTimeOut * 1000

        Return UdpSocket

    End Function

    ' Close down a Socket
    Private Sub Remove_Socket(ByVal Socket As Net.Sockets.Socket)
        Socket.Shutdown(Net.Sockets.SocketShutdown.Both)
        Socket.Close()
    End Sub

    Private Sub Button1_Click(sender As Object, e As EventArgs) Handles Button1.Click

        ' Create a Byte Array for the DHCPDISCOVER packet
        Dim Message As Byte() = New_DhcpDiscoverPacket()

        ' Create a socket
        Dim UdpSocket As Net.Sockets.Socket = New_UdpSocket(20, 20)

        ' UDP Port 68 (Server-to-Client port)
        Dim EndPoint As Net.EndPoint = New Net.IPEndPoint(Net.IPAddress.Any, 68)

        ' Listen on $EndPoint
        UdpSocket.Bind(EndPoint)

        ' UDP Port 67 (Client-to-Server port)
        EndPoint = New Net.IPEndPoint(Net.IPAddress.Broadcast, 67)
        ' Send the DHCPDISCOVER packet
        Dim BytesSent As Integer = UdpSocket.SendTo(Message, EndPoint)

        ' Begin receiving And processing responses
        Dim NoConnectionTimeOut As Boolean = True

        Dim Start As Date = Date.Now

        While NoConnectionTimeOut

            Dim BytesReceived As Integer = 0
            Dim ReceiveBuffer As Byte() = New Byte(1023) {}

            Try

                ' Placeholder EndPoint for the Sender
                Dim SenderEndPoint As Net.EndPoint = New Net.IPEndPoint(Net.IPAddress.Any, 0)
                ' Receive Buffer
                BytesReceived = UdpSocket.ReceiveFrom(ReceiveBuffer, SenderEndPoint)

            Catch ex As Net.Sockets.SocketException

                ' Catch a SocketException, thrown when the Receive TimeOut value is reached
                NoConnectionTimeOut = False
                Debug.Print("SocketExceptionTimeOut")

            End Try

            If BytesReceived > 0 Then
                Debug.Print("")
                Debug.Print("BytesReceived")
                Debug.Print("****************************************************")
                Call Read_DhcpPacket(ReceiveBuffer)
            End If

            If Date.Now > Start.AddSeconds(DiscoverTimeout) Then
                ' Exit condition, Not error condition
                NoConnectionTimeOut = False
                Debug.Print("DiscoverTimeout")
            End If

        End While


        Call Remove_Socket(UdpSocket)

        Debug.Print("Ready")
    End Sub

End Class
